#! /usr/local/bin/ost
"""Translate models from Tara/ Xabi from PDB + extra data into ModelCIF."""
# ToDo [internal]: get DB versions in - https://colabfold.mmseqs.com, scroll
# down to "Database Information"

import argparse
import datetime
import os
import sys

from timeit import default_timer as timer
import numpy as np
import requests
import ujson as json

import ihm
import modelcif
import modelcif.associated
import modelcif.dumper
import modelcif.model
import modelcif.protocol
import modelcif.reference

from ost import io


def _parse_args():
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description=__doc__,
    )

    parser.add_argument(
        "model_dir",
        type=str,
        metavar="<MODEL DIR>",
        help="Directory with model(s) to be translated. Must be of form "
        + "'<UniProtKB AC>-<UniProtKB AC>'",
    )

    opts = parser.parse_args()

    # check that model dir exists
    if opts.model_dir.endswith("/"):
        opts.model_dir = opts.model_dir[:-1]
    if not os.path.exists(opts.model_dir):
        _abort_msg(f"Model directory '{opts.model_dir}' does not exist.")
    if not os.path.isdir(opts.model_dir):
        _abort_msg(f"Path '{opts.model_dir}' does not point to a directory.")

    return opts


# pylint: disable=too-few-public-methods
class _GlobalPTM(modelcif.qa_metric.Global, modelcif.qa_metric.PTM):
    """Predicted accuracy according to the TM-score score in [0,1]."""

    name = "pTM"
    software = None


class _GlobalPLDDT(modelcif.qa_metric.Global, modelcif.qa_metric.PLDDT):
    """Predicted accuracy according to the CA-only lDDT in [0,100]."""

    name = "pLDDT"
    software = None


class _LocalPLDDT(modelcif.qa_metric.Local, modelcif.qa_metric.PLDDT):
    """Predicted accuracy according to the CA-only lDDT in [0,100]."""

    name = "pLDDT"
    software = None


class _PAE(modelcif.qa_metric.MetricType):
    """Predicted aligned error (in Angstroms).
    See :class:`MetricType` for more information."""

    type = "PAE"
    other_details = None


class _LocalPairwisePAE(modelcif.qa_metric.LocalPairwise, _PAE):
    """predicted aligned error (in Angstroms)."""

    name = "PAE"
    software = None


# pylint: enable=too-few-public-methods


class _OST2ModelCIF(modelcif.model.AbInitioModel):
    """Map OST entity elements to ihm.model"""

    def __init__(self, *args, **kwargs):
        """Initialise a model"""
        self.ost_entity = kwargs.pop("ost_entity")
        self.asym = kwargs.pop("asym")

        super().__init__(*args, **kwargs)

    def get_atoms(self):
        # ToDo [internal]: Take B-factor out since its not a B-factor?
        for atm in self.ost_entity.atoms:
            yield modelcif.model.Atom(
                asym_unit=self.asym[atm.chain.name],
                seq_id=atm.residue.number.num,
                atom_id=atm.name,
                type_symbol=atm.element,
                x=atm.pos[0],
                y=atm.pos[1],
                z=atm.pos[2],
                het=atm.is_hetatom,
                biso=atm.b_factor,
                occupancy=atm.occupancy,
            )

    def add_scores(self, scores_json, entry_id, ac_file_prfx):
        """Add QA metrics from AF2 scores."""
        # global scores
        self.qa_metrics.extend(
            (
                _GlobalPLDDT(np.mean(scores_json["plddt"])),
                _GlobalPTM(scores_json["ptm"]),
            )
        )

        # local scores
        lpae = []
        i = 0
        for chn_i in self.ost_entity.chains:
            for res_i in chn_i.residues:
                # local pLDDT
                self.qa_metrics.append(
                    _LocalPLDDT(
                        self.asym[chn_i.name].residue(res_i.number.num),
                        scores_json["plddt"][i],
                    )
                )

                # pairwise alignment error
                j = 0
                # We do a 2nd iteration over the OST entity above doing index
                # magic because it keeps the code cleaner and is only around
                # 0.3s slower than iterating the array directly. Majority of
                # time goes into writing files, anyway.
                for chn_j in self.ost_entity.chains:
                    for res_j in chn_j.residues:
                        lpae.append(
                            _LocalPairwisePAE(
                                self.asym[chn_i.name].residue(res_i.number.num),
                                self.asym[chn_j.name].residue(res_j.number.num),
                                scores_json["pae"][i][j],
                            )
                        )
                        j += 1

                i += 1

        self.qa_metrics.extend(lpae)

        ac_file = f"{os.path.basename(ac_file_prfx)}_local_pairwise_qa.cif"

        return modelcif.associated.Repository(
            "",
            [
                modelcif.associated.LocalPairwiseQAScoresFile(
                    ac_file,
                    categories=["_ma_qa_metric_local_pairwise"],
                    copy_categories=["_ma_qa_metric"],
                    entry_id=entry_id,
                    entry_details="This file is an associated file consisting "
                    + "of local pairwise QA metrics. This is a partial mmCIF "
                    + "file and can be validated by merging with the main "
                    + "mmCIF file containing the model coordinates and other "
                    + "associated data.",
                    details="Predicted aligned error.",
                )
            ],
        )


def _abort_msg(msg, exit_code=1):
    """Write error message and exit with exit_code."""
    print(f"{msg}\nAborting.", file=sys.stderr)
    sys.exit(exit_code)


def _check_file(file_path):
    """Make sure a file exists and is actually a file."""
    if not os.path.exists(file_path):
        _abort_msg(f"File not found: '{file_path}'.")
    if not os.path.isfile(file_path):
        _abort_msg(f"File path does not point to file: '{file_path}'.")


def _check_interaction_extra_files_present(model_dir):
    """Make sure some general files are present for an interaction."""
    cnfg = os.path.join(model_dir, "config.json")
    _check_file(cnfg)

    return cnfg


def _check_model_extra_files_present(model_dir, pdb_file):
    """Check that all files needed to process this model are present."""
    uid = os.path.splitext(pdb_file)[0]
    prfx = os.path.join(model_dir, uid)

    scrs = f"{prfx}_scores.json"
    _check_file(scrs)

    return prfx, uid


def _get_audit_authors():
    """Return the list of authors that produced this model."""
    # ToDo: tell Xabi that his name can't have a á in mmCIF
    return (
        "Bartolec T",
        "Vazquez-Campos X",
        "Johnson M",
        "Norman A",
        "Payne R",
        "Wilkins M",
        "Mackay J",
        "Low J",
    )


def _get_protocol_steps_and_software(cnfg_file):
    """Create the list of protocol steps with software and parameters used."""
    protocol = []

    # modelling step
    step = {
        "method_type": "modeling",
        "name": "ma_protocol_step.step_name",
        "details": "Model using AlphaFold-Multimer (AlphaFold v2.2.0), "
        + "without amber relaxation and producing 5 models with up to 3 "
        + "recycles each, starting from paired and unparied MSAs for the "
        + "dimers using MMseqs2.",
    }
    # get input data
    # Must refer to data already in the JSON, so we try keywords
    step["input"] = "target_sequences"
    # get output data
    # Must refer to existing data, so we try keywords
    step["output"] = "model"
    # get software
    step["software"] = [
        {
            "name": "ColabFold",
            "classification": "model building",
            # ToDo: Get description for ColabFold
            "description": "software.description",
            "citation": {
                "pmid": None,
                "title": "ColabFold - Making protein folding accessible to all",
                "journal": "bioRxiv",
                "volume": None,
                "page_range": None,
                "year": 2022,
                "authors": [
                    "Mirdita M",
                    "Schuetze K",
                    "Moriwaki Y",
                    "Heo L",
                    "Ovchinnikov S",
                    "Steinegger M",
                ],
                "doi": "10.1101/2021.08.15.456425",
            },
            "location": "https://github.com/sokrypton/ColabFold",
            "type": "package",
            "version": "1.2.0",
        },
        {
            "name": "MMseqs2",
            "classification": "data collection",
            "description": "Many-against-Many sequence searching",
            "citation": {
                "pmid": "30615063",
                "title": "MMseqs2 desktop and local web server app for fast, "
                + "interactive sequence searches",
                "journal": "Bioinformatics",
                "volume": 35,
                "page_range": (2856, 2858),
                "year": 2019,
                "authors": [
                    "Mirdita M",
                    "Steinegger M",
                    "Soeding J",
                ],
                "doi": "10.1093/bioinformatics/bty1057",
            },
            "location": "https://github.com/soedinglab/mmseqs2",
            "type": "package",
            "version": None,
        },
        {
            "name": "AlphaFold-Multimer",
            "classification": "model building",
            "description": "Structure prediction",
            "citation": {
                "pmid": None,
                "title": "Protein complex prediction with "
                + "AlphaFold-Multimer.",
                "journal": "bioRxiv",
                "volume": None,
                "page_range": None,
                "year": 2021,
                "authors": [
                    "Evans R",
                    "O'Neill M",
                    "Pritzel A",
                    "Antropova N",
                    "Senior A",
                    "Green T",
                    "Zidek A",
                    "Bates R",
                    "Blackwell S",
                    "Yim J",
                    "Ronneberger O",
                    "Bodenstein S",
                    "Zielinski M",
                    "Bridgland A",
                    "Potapenko A",
                    "Cowie A",
                    "Tunyasuvunakool K",
                    "Jain R",
                    "Clancy E",
                    "Kohli P",
                    "Jumper J",
                    "Hassabis D",
                ],
                "doi": "10.1101/2021.10.04.463034",
            },
            "location": "https://github.com/deepmind/alphafold",
            "type": "package",
            "version": "2.1.1",
        },
    ]
    # get parameters
    with open(cnfg_file, encoding="utf8") as jfh:
        step["software_parameters"] = json.load(jfh)
    protocol.append(step)

    # model selection step
    # ToDo [input/ internal]: model selection step on a single model is a bit
    # silly, how do we get a list of models?
    step = {
        "method_type": "model selection",
        "name": "ma_protocol_step.step_name",
        "details": "Select best model, which is either the top-ranked model "
        + "as determined by the ColabFold pipeline "
        + "(iptmscore*0.8+ptmscore*0.2), or else the model with best "
        + "congruence with crosslinks reported in the related study.",
    }
    step["input"] = "model"
    step["output"] = "model"
    step["software"] = []
    step["software_parameters"] = {}
    protocol.append(step)

    return protocol


def _get_title(gene_names):
    """Get a title for this modelling experiment."""
    return f"Predicted interaction between {' and '.join(gene_names)}"


def _get_model_details(gene_names):
    """Get the model description."""
    return (
        f"Dimer model generated for {' and '.join(gene_names)}, produced "
        + "using AlphaFold-Multimer (AlphaFold v2.2.0) as implemented by "
        + "ColabFold (v1.2.0) which uses MMseqs2 for MSA generation (UniRef30 "
        + "+ Environmental)."
    )


def _get_model_group_name():
    """Get a name for a model group."""

    return "Crosslinked Heterodimer ALphaFold-Multimer v2 Models"


def _get_sequence(chn):
    """Get the sequence out of an OST chain."""
    # initialise
    lst_rn = chn.residues[0].number.num
    idx = 1
    sqe = chn.residues[0].one_letter_code
    if lst_rn != 1:
        sqe = "-"
        idx = 0

    for res in chn.residues[idx:]:
        lst_rn += 1
        while lst_rn != res.number.num:
            sqe += "-"
            lst_rn += 1
        sqe += res.one_letter_code

    return sqe


def _check_sequence(up_ac, sequence):
    """Verify sequence to only contain standard olc."""
    for res in sequence:
        if res not in "ACDEFGHIKLMNPQRSTVWY":
            raise RuntimeError(
                "Non-standard aa found in UniProtKB sequence "
                + f"for entry '{up_ac}': {res}"
            )


def _fetch_upkb_entry(up_ac):
    """Fetch data for an UniProtKB entry."""
    # This is a simple parser for UniProtKB txt format, instead of breaking it up
    # into multiple functions, we just allow many many branches & statements,
    # here.
    # pylint: disable=too-many-branches,too-many-statements
    data = {}
    data["up_organism"] = ""
    data["up_sequence"] = ""
    data["up_ac"] = up_ac
    rspns = requests.get(f"https://www.uniprot.org/uniprot/{up_ac}.txt")
    for line in rspns.iter_lines(decode_unicode=True):
        if line.startswith("ID   "):
            sline = line.split()
            if len(sline) != 5:
                _abort_msg(f"Unusual UniProtKB ID line found:\n'{line}'")
            data["up_id"] = sline[1]
        elif line.startswith("OX   NCBI_TaxID="):
            # Following strictly the UniProtKB format: 'OX   NCBI_TaxID=<ID>;'
            data["up_ncbi_taxid"] = line[len("OX   NCBI_TaxID=") : -1]
            data["up_ncbi_taxid"] = data["up_ncbi_taxid"].split("{")[0].strip()
        elif line.startswith("OS   "):
            if line[-1] == ".":
                data["up_organism"] += line[len("OS   ") : -1]
            else:
                data["up_organism"] += line[len("OS   ") : -1] + " "
        elif line.startswith("SQ   "):
            sline = line.split()
            if len(sline) != 8:
                _abort_msg(f"Unusual UniProtKB SQ line found:\n'{line}'")
            data["up_seqlen"] = int(sline[2])
            data["up_crc64"] = sline[6]
        elif line.startswith("     "):
            sline = line.split()
            if len(sline) > 6:
                _abort_msg(
                    "Unusual UniProtKB sequence data line "
                    + f"found:\n'{line}'"
                )
            data["up_sequence"] += "".join(sline)
        elif line.startswith("RP   "):
            if "ISOFORM" in line.upper():
                RuntimeError(
                    f"First ISOFORM found for '{up_ac}', needs " + "handling."
                )
        elif line.startswith("DT   "):
            # 2012-10-03
            dt_flds = line[len("DT   ") :].split(", ")
            if dt_flds[1].upper().startswith("SEQUENCE VERSION "):
                data["up_last_mod"] = datetime.datetime.strptime(
                    dt_flds[0], "%d-%b-%Y"
                )
        elif line.startswith("GN   Name="):
            data["up_gn"] = line[len("GN   Name=") :].split(";")[0]
            data["up_gn"] = data["up_gn"].split("{")[0].strip()

    # we have not seen isoforms in the data set, yet, so we just set them to '.'
    data["up_isoform"] = None

    if "up_gn" not in data:
        _abort_msg(f"No gene name found for UniProtKB entry '{up_ac}'.")
    if "up_last_mod" not in data:
        _abort_msg(f"No sequence version found for UniProtKB entry '{up_ac}'.")
    if "up_crc64" not in data:
        _abort_msg(f"No CRC64 value found for UniProtKB entry '{up_ac}'.")
    if len(data["up_sequence"]) == 0:
        _abort_msg(f"No sequence found for UniProtKB entry '{up_ac}'.")
    # check that sequence length and CRC64 is correct
    if data["up_seqlen"] != len(data["up_sequence"]):
        _abort_msg(
            "Sequence length of SQ line and sequence data differ for "
            + f"UniProtKB entry '{up_ac}': {data['up_seqlen']} != "
            + f"{len(data['up_sequence'])}"
        )
    _check_sequence(data["up_ac"], data["up_sequence"])

    if "up_id" not in data:
        _abort_msg(f"No ID found for UniProtKB entry '{up_ac}'.")
    if "up_ncbi_taxid" not in data:
        _abort_msg(f"No NCBI taxonomy ID found for UniProtKB entry '{up_ac}'.")
    if len(data["up_organism"]) == 0:
        _abort_msg(f"No organism species found for UniProtKB entry '{up_ac}'.")

    return data


def _get_upkb_for_sequence(sqe, up_ac):
    """Get UniProtKB entry data for given sequence."""
    up_data = _fetch_upkb_entry(up_ac)
    if sqe != up_data["up_sequence"]:
        raise RuntimeError(
            f"Sequences not equal from file: {sqe}, from UniProtKB: "
            + f"{up_data['up_sequence']}"
        )

    return up_data


def _get_entities(pdb_file, up_acs):
    """Gather data for the mmCIF (target) entities."""
    entities = []

    ost_ent = io.LoadPDB(pdb_file)
    for i, chn in enumerate(ost_ent.chains):
        cif_ent = {}
        sqe = _get_sequence(chn)
        upkb = _get_upkb_for_sequence(sqe, up_acs[i])
        cif_ent["pdb_sequence"] = sqe
        cif_ent["pdb_chain_id"] = chn.name
        cif_ent["description"] = f"Model of {upkb['up_gn']} ({upkb['up_ac']})"
        cif_ent.update(upkb)
        entities.append(cif_ent)

    return entities, ost_ent


def _get_scores(data, prfx):
    """Check that all files needed to process this model are present."""
    scrs_fle = f"{prfx}_scores.json"
    with open(scrs_fle, encoding="utf8") as jfh:
        scrs_json = json.load(jfh)

    # ToDo: is dict.update still the way to go when iterating multiple model
    # directories? Aka, does dict.update overwrite old scores?
    data.update(scrs_json)


def _get_modelcif_entities(target_ents, source, asym_units, system):
    """Create ModelCIF entities and asymmetric units."""
    for cif_ent in target_ents:
        mdlcif_ent = modelcif.Entity(
            cif_ent["pdb_sequence"],
            description=cif_ent["description"],
            source=source,
            references=[
                modelcif.reference.UniProt(
                    cif_ent["up_id"],
                    cif_ent["up_ac"],
                    # Careful: alignments are just that easy because of equal
                    # sequences!
                    align_begin=1,
                    align_end=cif_ent["up_seqlen"],
                    isoform=cif_ent["up_isoform"],
                    ncbi_taxonomy_id=cif_ent["up_ncbi_taxid"],
                    organism_scientific=cif_ent["up_organism"],
                    sequence_version_date=cif_ent["up_last_mod"],
                    sequence_crc64=cif_ent["up_crc64"],
                )
            ],
        )
        # ToDo [input]: Add details
        asym_units[cif_ent["pdb_chain_id"]] = modelcif.AsymUnit(
            mdlcif_ent,
            details="struct_asym.details",
        )
        system.target_entities.append(mdlcif_ent)


def _assemble_modelcif_software(soft_dict):
    """Create a modelcif.Software instance from dictionary."""
    return modelcif.Software(
        soft_dict["name"],
        soft_dict["classification"],
        soft_dict["description"],
        soft_dict["location"],
        soft_dict["type"],
        soft_dict["version"],
        citation=ihm.Citation(
            pmid=soft_dict["citation"]["pmid"],
            title=soft_dict["citation"]["title"],
            journal=soft_dict["citation"]["journal"],
            volume=soft_dict["citation"]["volume"],
            page_range=soft_dict["citation"]["page_range"],
            year=soft_dict["citation"]["year"],
            authors=soft_dict["citation"]["authors"],
            doi=soft_dict["citation"]["doi"],
        ),
    )


def _get_modelcif_protocol(protocol_steps, target_entities, model):
    """Create the protocol for the ModelCIF file."""
    protocol = modelcif.protocol.Protocol()
    for js_step in protocol_steps:
        sftwre = None
        # ToDo [input]: Turn into software group if parameters are available
        if js_step["software"]:
            if len(js_step["software"]) == 1:
                sftwre = _assemble_modelcif_software(js_step["software"][0])
            else:
                sftwre = []
                for sft in js_step["software"]:
                    sftwre.append(_assemble_modelcif_software(sft))
                sftwre = modelcif.SoftwareGroup(elements=sftwre)
            if js_step["software_parameters"]:
                params = []
                # ToDo [internal]: handle lists!
                for k, v in js_step["software_parameters"].items():
                    params.append(
                        modelcif.SoftwareParameter(k, v)
                    )
                if isinstance(sftwre, modelcif.SoftwareGroup):
                    sftwre.parameters = params
                else:
                    sftwre = modelcif.SoftwareGroup(
                        elements=(sftwre,), parameters=params
                    )

        if js_step["input"] == "target_sequences":
            input_data = modelcif.data.DataGroup(target_entities)
        # ToDo: Add databases + versions
        elif js_step["input"] == "model":
            input_data = model
        else:
            raise RuntimeError(f"Unknown protocol input: '{js_step['input']}'")
        if js_step["output"] == "model":
            output_data = model
        else:
            raise RuntimeError(
                f"Unknown protocol output: '{js_step['output']}'"
            )
        protocol.steps.append(
            modelcif.protocol.Step(
                input_data=input_data,
                output_data=output_data,
                name=js_step["name"],
                details=js_step["details"],
                software=sftwre,
            )
        )
        protocol.steps[-1].method_type = js_step["method_type"]

    return protocol


def _store_as_modelcif(interaction_name, data_json, ost_ent, file_prfx):
    """Mix all the data into a ModelCIF file."""
    print("    generating ModelCIF objects...", end="")
    pstart = timer()
    # create system to gather all the data
    system = modelcif.System(
        title=data_json["title"],
        id=interaction_name.upper(),
        model_details=data_json["model_details"],
    )
    # create target entities, references, source, asymmetric units & assembly
    # for source we assume all chains come from the same taxon
    source = ihm.source.Natural(
        ncbi_taxonomy_id=data_json["target_entities"][0]["up_ncbi_taxid"],
        scientific_name=data_json["target_entities"][0]["up_organism"],
    )

    # create an asymmetric unit and an entity per target sequence
    asym_units = {}
    _get_modelcif_entities(
        data_json["target_entities"], source, asym_units, system
    )

    # ToDo [input]: Get Assembly name
    assembly = modelcif.Assembly(
        asym_units.values(), name="ma_struct_assembly_details.assembly_name"
    )

    # audit_authors
    system.authors.extend(data_json["audit_authors"])

    # set up the model to produce coordinates
    # ToDo [input]: Get ma_model_list.model_name
    model = _OST2ModelCIF(
        assembly=assembly,
        asym=asym_units,
        ost_entity=ost_ent,
        name="ma_model_list.model_name",
    )
    print(f" ({timer()-pstart:.2f}s)")
    print("    processing QA scores...", end="", flush=True)
    pstart = timer()
    system.repositories.append(
        model.add_scores(data_json, system.id, file_prfx)
    )
    print(f" ({timer()-pstart:.2f}s)")

    model_group = modelcif.model.ModelGroup(
        [model], name=data_json["model_group_name"]
    )
    system.model_groups.append(model_group)

    protocol = _get_modelcif_protocol(
        data_json["protocol"], system.target_entities, model
    )
    system.protocols.append(protocol)

    # write modelcif System to file
    print("    write to disk...", end="", flush=True)
    pstart = timer()
    with open(f"{file_prfx}.cif", "w", encoding="ascii") as mmcif_fh:
        modelcif.dumper.write(mmcif_fh, [system])
    print(f" ({timer()-pstart:.2f}s)")


def _create_interaction_json(cnfg_file):
    """Create a dictionary (mimicking JSON) that contains data which is the same
    for all models."""
    data = {}

    data["audit_authors"] = _get_audit_authors()
    data["protocol"] = _get_protocol_steps_and_software(cnfg_file)

    return data


def _create_model_json(data, pdb_file, up_acs):
    """Create a dictionary (mimicking JSON) that contains all the data."""

    data["target_entities"], ost_ent = _get_entities(pdb_file, up_acs)
    gns = []
    for i in data["target_entities"]:
        gns.append(i["up_gn"])
    data["title"] = _get_title(gns)
    data["model_details"] = _get_model_details(gns)
    data["model_group_name"] = _get_model_group_name()

    return ost_ent


def _main():
    """Run as script."""
    opts = _parse_args()

    interaction = os.path.split(opts.model_dir)[-1]
    print(f"Working on {interaction}...")

    # get UniProtKB ACs from directory name
    up_acs = interaction.split("-")

    cnfg = _check_interaction_extra_files_present(opts.model_dir)

    mdlcf_json = _create_interaction_json(cnfg)

    # iterate model directory
    for fle in os.listdir(opts.model_dir):
        # iterate PDB files
        if not fle.endswith(".pdb"):
            continue
        print(f"  translating {fle}...")
        pdb_start = timer()
        file_prfx, uid = _check_model_extra_files_present(opts.model_dir, fle)
        fle = os.path.join(opts.model_dir, fle)

        # gather data into JSON-like structure
        print("    preparing data...", end="")
        pstart = timer()
        ost_ent = _create_model_json(mdlcf_json, fle, up_acs)

        # read quality scores from JSON file
        _get_scores(mdlcf_json, file_prfx)
        print(f" ({timer()-pstart:.2f}s)")

        _store_as_modelcif(uid, mdlcf_json, ost_ent, file_prfx)
        # ToDo [internal]: wipe data or is it overwritten in mdlcf_json?
        print(f"  ... done with {fle} ({timer()-pdb_start:.2f}s).")

    print(f"... done with {opts.model_dir}.")


if __name__ == "__main__":
    _main()

#  LocalWords:  Xabi argparse args ArgumentParser formatter dir str metavar os
#  LocalWords:  RawDescriptionHelpFormatter ACs sys msg nAborting stderr acs
#  LocalWords:  fle cnfg config json pdb prfx scrs endswith modelcif uid pdbx
#  LocalWords:  struct cif utf mmcif fh datetime ost io ToDo chn lst rn idx aa
#  LocalWords:  sqe olc ACDEFGHIKLMNPQRSTVWY RuntimeError upkb txt pylint iter
#  LocalWords:  rspns unicode startswith sline len elif NCBI TaxID ncbi taxid
#  LocalWords:  seqlen crc ISOFORM DT dt flds isoforms isoform ent LoadPDB ihm
#  LocalWords:  mdlcf mdlcif asym AsymUnit init kwargs atm pos het hetatom pTM
#  LocalWords:  biso ujson GlobalPTM pLDDT ptm jfh numpy np GlobalPLDDT lDDT
#  LocalWords:  plddt LocalPLDDT timeit PAE MetricType LocalPairwisePAE lpae
#  LocalWords:  nd pae qa pstart ColabFold
